# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
import typing
from typing import Any, Callable, Dict, List, Optional, Tuple, cast
import math
import cmath
import random
import warnings

import numpy as np

from ..gate import *
from ..utils import ignore_global_phase
from .backendbase import Backend

DEFAULT_DTYPE = complex

OPS_DISABLE_CTX_CACHE = (Measurement, Reset)

DEFAULT_SHOTS: int = 1024


class _NumPyBackendContext:
    """This class is internally used in NumPyBackend"""
    def __init__(self,
                 n_qubits: int,
                 cache: Optional[np.ndarray] = None,
                 cache_idx: int = -1) -> None:
        self.n_qubits = n_qubits
        self.qubits = np.zeros(2**n_qubits, dtype=DEFAULT_DTYPE)
        self.qubits_buf = np.zeros(2**n_qubits, dtype=DEFAULT_DTYPE)
        self.indices = np.arange(2**n_qubits, dtype=np.uint32)
        self.save_ctx_cache = True
        self.cache = cache
        self.cache_idx = cache_idx
        self.shots_result = Counter()
        self.cregs = [0] * self.n_qubits

    def prepare(self, initial: Optional[np.ndarray]) -> None:
        """Prepare to run next shot."""
        if self.cache is not None:
            np.copyto(self.qubits, self.cache)
        elif initial is not None:
            np.copyto(self.qubits, initial)
        else:
            self.qubits.fill(0.0)
            self.qubits[0] = 1.0
        self.cregs = [0] * self.n_qubits
        self.sample = {}

    def store_shot(self) -> None:
        """Store current cregs to shots_result"""
        def to_str(cregs: List[int]) -> str:
            return ''.join(str(b) for b in cregs)

        key = to_str(self.cregs)
        self.shots_result[key] = self.shots_result.get(key, 0) + 1


class _NumPyBackendOperations:
    @staticmethod
    def gate_x(gate: XGate, ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of X gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = qubits[t1]
            newq[t1] = qubits[t0]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_y(gate: YGate, ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of Y gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = -1.0j * qubits[t1]
            newq[t1] = 1.0j * qubits[t0]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_z(gate: ZGate, ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of Z gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= -1
        return ctx

    @staticmethod
    def gate_h(gate: HGate, ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of H gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = qubits[t0] + qubits[t1]
            newq[t1] = qubits[t0] - qubits[t1]
            newq *= 1.0 / math.sqrt(2)
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_rx(gate: RXGate,
                ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of RX gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = math.cos(halftheta)
        a01 = a10 = -1j * math.sin(halftheta)
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = a00 * qubits[t0] + a01 * qubits[t1]
            newq[t1] = a10 * qubits[t0] + a11 * qubits[t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_ry(gate: RYGate,
                ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of RY gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = math.cos(halftheta)
        a10 = math.sin(halftheta)
        a01 = -a10
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = a00 * qubits[t0] + a01 * qubits[t1]
            newq[t1] = a10 * qubits[t0] + a11 * qubits[t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_rz(gate: RZGate,
                ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of RZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a0 = complex(math.cos(halftheta), -math.sin(halftheta))
        a1 = complex(math.cos(halftheta), math.sin(halftheta))
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) == 0] *= a0
            qubits[(i & (1 << target)) != 0] *= a1
        return ctx

    @staticmethod
    def gate_phase(gate: PhaseGate,
                   ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of Phase gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        a = complex(math.cos(theta), math.sin(theta))
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= a
        return ctx

    @staticmethod
    def gate_t(gate: PhaseGate,
               ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of T gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices

        sqrt2_inv = 1 / math.sqrt(2)
        factor = complex(sqrt2_inv, sqrt2_inv)
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= factor
        return ctx

    @staticmethod
    def gate_s(gate: SGate, ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of S gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= 1.j
        return ctx

    @staticmethod
    def gate_cz(gate: CZGate,
                ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for control, target in gate.control_target_iter(n_qubits):
            qubits[((i & (1 << control)) != 0)
                   & ((i & (1 << target)) != 0)] *= -1
        return ctx

    @staticmethod
    def gate_cx(gate: CXGate,
                ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CX gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[c1 & t0] = qubits[c1 & t1]
            newq[c1 & t1] = qubits[c1 & t0]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_cu(gate: CUGate,
                ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CU gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        phi = gate.phi
        lam = gate.lam
        gamma = gate.gamma
        globalphase = cmath.exp(1j * gamma)
        a00 = math.cos(theta * 0.5) * globalphase
        a11 = a00 * cmath.exp(1j * (phi + lam))
        a01 = a10 = math.sin(theta * 0.5) * globalphase
        a01 *= -cmath.exp(1j * lam)
        a10 *= cmath.exp(1j * phi)
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            c1t0 = ((i & (1 << target)) == 0) & c1
            c1t1 = ((i & (1 << target)) != 0) & c1
            newq[c1t0] = a00 * qubits[c1t0] + a01 * qubits[c1t1]
            newq[c1t1] = a10 * qubits[c1t0] + a11 * qubits[c1t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_crx(gate: CRXGate,
                 ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CRX gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = math.cos(halftheta)
        a01 = a10 = -1j * math.sin(halftheta)
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            c1t0 = ((i & (1 << target)) == 0) & c1
            c1t1 = ((i & (1 << target)) != 0) & c1
            newq[c1t0] = a00 * qubits[c1t0] + a01 * qubits[c1t1]
            newq[c1t1] = a10 * qubits[c1t0] + a11 * qubits[c1t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_cry(gate: CRYGate,
                 ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CRY gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = math.cos(halftheta)
        a10 = math.sin(halftheta)
        a01 = -a10
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            c1t0 = ((i & (1 << target)) == 0) & c1
            c1t1 = ((i & (1 << target)) != 0) & c1
            newq[c1t0] = a00 * qubits[c1t0] + a01 * qubits[c1t1]
            newq[c1t1] = a10 * qubits[c1t0] + a11 * qubits[c1t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_crz(gate: CRZGate,
                 ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CRZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a0 = complex(math.cos(halftheta), -math.sin(halftheta))
        a1 = complex(math.cos(halftheta), math.sin(halftheta))
        for control, target in gate.control_target_iter(n_qubits):
            c1t0 = ((i & (1 << control)) != 0) & ((i & (1 << target)) == 0)
            c1t1 = ((i & (1 << control)) != 0) & ((i & (1 << target)) != 0)
            qubits[c1t0] *= a0
            qubits[c1t1] *= a1
        return ctx

    @staticmethod
    def gate_cphase(gate: CPhaseGate,
                    ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CPhase gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        a = complex(math.cos(theta), math.sin(theta))
        for control, target in gate.control_target_iter(n_qubits):
            c1t1 = ((i & (1 << control)) != 0) & ((i & (1 << target)) != 0)
            qubits[c1t1] *= a
        return ctx

    @staticmethod
    def gate_ccz(gate: CCZGate,
                 ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of CCZ gate."""
        c1, c2, t = cast(Tuple[int, int, int], gate.targets)
        qubits = ctx.qubits
        i = ctx.indices
        indices = (i & (1 << c1)) != 0
        indices &= (i & (1 << c2)) != 0
        indices &= (i & (1 << t)) != 0
        qubits[indices] *= -1
        return ctx

    @staticmethod
    def gate_ccx(gate: ToffoliGate,
                 ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of Toffoli gate."""
        _, _, t = cast(Tuple[int, int, int], gate.targets)
        ctx = _NumPyBackendOperations.gate_h(HGate(t), ctx)
        ctx = _NumPyBackendOperations.gate_ccz(CCZGate(gate.targets), ctx)
        ctx = _NumPyBackendOperations.gate_h(HGate(t), ctx)
        return ctx

    @staticmethod
    def gate_u(gate: UGate, ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of U gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        phi = gate.phi
        lam = gate.lam
        gamma = gate.gamma
        globalphase = cmath.exp(1j * gamma)
        a00 = math.cos(theta * 0.5) * globalphase
        a11 = a00 * cmath.exp(1j * (phi + lam))
        a01 = a10 = math.sin(theta * 0.5) * globalphase
        a01 *= -cmath.exp(1j * lam)
        a10 *= cmath.exp(1j * phi)
        for target in gate.target_iter(n_qubits):
            np.copyto(newq, qubits)
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = qubits[t0] * a00
            newq[t0] += qubits[t1] * a01
            newq[t1] = qubits[t0] * a10
            newq[t1] += qubits[t1] * a11
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_mat1(gate: Mat1Gate,
                  ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of U3 gate."""
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        mat = gate.mat
        for target in gate.target_iter(n_qubits):
            np.copyto(newq, qubits)
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = mat[0, 0] * qubits[t0]
            newq[t0] += mat[0, 1] * qubits[t1]
            newq[t1] = mat[1, 0] * qubits[t0]
            newq[t1] += mat[1, 1] * qubits[t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    @staticmethod
    def gate_measure(gate: Measurement,
                     ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of measurement operation."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        measured = []
        for target in gate.target_iter(n_qubits):
            p_zero = np.linalg.norm(qubits[(i & (1 << target)) == 0])**2
            rand = random.random()
            if rand < p_zero:
                qubits[(i & (1 << target)) != 0] = 0.0
                qubits /= math.sqrt(p_zero)
                ctx.cregs[target] = 0
                measured.append(0)
            else:
                qubits[(i & (1 << target)) == 0] = 0.0
                qubits /= math.sqrt(1.0 - p_zero)
                ctx.cregs[target] = 1
                measured.append(1)
        if gate.key is not None:
            if gate.key in ctx.sample:
                if gate.duplicated == "replace":
                    ctx.sample[gate.key] = measured
                elif gate.duplicated == "append":
                    ctx.sample[gate.key] += measured
                else:
                    raise ValueError("Measurement key is duplicated.")
            else:
                ctx.sample[gate.key] = measured
        return ctx

    @staticmethod
    def gate_reset(gate: Reset,
                   ctx: _NumPyBackendContext) -> _NumPyBackendContext:
        """Implementation of measurement operation."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            p_zero = np.linalg.norm(qubits[(i & (1 << target)) == 0])**2
            rand = random.random()
            t1 = (i & (1 << target)) != 0
            if rand < p_zero:
                qubits[t1] = 0.0
                qubits /= math.sqrt(p_zero)
            else:
                qubits[(i & (1 << target)) == 0] = qubits[t1]
                qubits[t1] = 0.0
                qubits /= math.sqrt(1.0 - p_zero)
        return ctx


def _check_and_transform_initial(initial: np.ndarray,
                                 n_qubits: int) -> np.ndarray:
    """Check the shape and transform the type of initial."""
    if not isinstance(initial, np.ndarray):
        raise ValueError(
            f"`initial` must be a np.ndarray, but {type(initial)}")
    if initial.shape != (2**n_qubits, ):
        raise ValueError(
            f"`initial.shape` is not matched. Expected: {(2**n_qubits,)}, Actual: {initial.shape}"
        )
    if initial.dtype != DEFAULT_DTYPE:
        initial = initial.astype(DEFAULT_DTYPE)
    return initial


class NumPyBackend(Backend):
    """Simulator backend which uses numpy. This backend is Blueqat's default backend."""
    __return_type: Dict[str, Callable[[_NumPyBackendContext], Any]] = {
        "statevector": lambda ctx: ctx.qubits,
        "shots": lambda ctx: ctx.shots_result,
        "statevector_and_shots": lambda ctx: (ctx.qubits, ctx.shots_result),
        "_inner_ctx": lambda ctx: ctx,
        "samples": lambda _: _, # dummy
    }

    def __init__(self) -> None:
        self.cache = None
        self.cache_idx = -1

    def __clear_cache(self) -> None:
        self.cache = None
        self.cache_idx = -1

    def __clear_cache_if_invalid(self, n_qubits: int, dtype: type) -> None:
        if self.cache is None:
            self.__clear_cache()
            return
        if len(self.cache) != 2**n_qubits:
            self.__clear_cache()
            return
        if self.cache.dtype != dtype:
            self.__clear_cache()
            return

    @staticmethod
    def _run_inner(ctx: _NumPyBackendContext, gates: List[Operation],
                   n_qubits: int,
                   initial: Optional[np.ndarray]) -> _NumPyBackendContext:
        def run_single_gate(ctx: _NumPyBackendContext,
                            gate: Operation) -> _NumPyBackendContext:
            try:
                action = getattr(_NumPyBackendOperations,
                                 'gate_' + gate.lowername)
                ctx = action(gate, ctx)
            except AttributeError:
                if isinstance(gate, IFallbackOperation):
                    for g in gate.fallback(n_qubits):
                        ctx = run_single_gate(ctx, g)
                else:
                    raise ValueError(f"Unknown operation {gate.lowername}.")  # pylint: disable=raise-missing-from
            return ctx

        ctx.prepare(initial)
        for i, gate in enumerate(gates[ctx.cache_idx + 1:]):
            if isinstance(gate, OPS_DISABLE_CTX_CACHE):
                if ctx.save_ctx_cache:
                    ctx.cache = ctx.qubits.copy()
                    ctx.cache_idx += i
                    ctx.save_ctx_cache = False
            ctx = run_single_gate(ctx, gate)
        return ctx

    def run(self,
            gates: List[Operation],
            n_qubits,
            shots: Optional[int] = None,
            initial: Optional[np.ndarray] = None,
            returns: Optional[str] = None,
            ignore_global: bool = False,
            save_cache: bool = False,
            **kwargs) -> Any:
        def __parse_run_args(shots: Optional[int],
                             returns: Optional[str]) -> Tuple[int, str]:
            if returns is None:
                if shots is None:
                    returns = "statevector"
                else:
                    returns = "shots"
            if returns not in self.__return_type.keys():
                raise ValueError(f"Unknown returns type '{returns}'")
            if shots is None:
                if returns in ("statevector", "_inner_ctx"):
                    shots = 1
                else:
                    shots = DEFAULT_SHOTS
            if returns == "statevector" and shots > 1:
                shots = 1
                warnings.warn(
                    "When `returns` = 'statevector', `shots` is ignored.")
            return shots, returns

        shots, returns = __parse_run_args(shots, returns)
        if kwargs:
            warnings.warn(f"Unknown run arguments: {kwargs}")
        if returns == "samples":
            # Remarks: This feature is experimental.
            return self.samples(gates, n_qubits, shots, initial)

        if initial is not None:
            initial = _check_and_transform_initial(initial, n_qubits)
            if save_cache:
                warnings.warn(
                    "When initial is not None, saving cache is disabled.")
                save_cache = False
            self.__clear_cache()
        else:
            self.__clear_cache_if_invalid(n_qubits, DEFAULT_DTYPE)

        ctx = _NumPyBackendContext(n_qubits, self.cache, self.cache_idx)

        for _ in range(shots):
            ctx = NumPyBackend._run_inner(ctx, gates, n_qubits, initial)
            if ctx.cregs:
                ctx.store_shot()
            if save_cache:
                self.cache = ctx.cache
                self.cache_idx = ctx.cache_idx

        if ignore_global:
            ignore_global_phase(ctx.qubits)
        return self.__return_type[returns](ctx)

    @staticmethod
    def statevector(gates: List[Operation],
                    n_qubits,
                    initial: Optional[np.ndarray] = None) -> np.ndarray:
        if initial is not None:
            initial = _check_and_transform_initial(initial, n_qubits)
        ctx = _NumPyBackendContext(n_qubits)
        ctx = NumPyBackend._run_inner(ctx, gates, n_qubits, initial)
        return ctx.qubits

    @staticmethod
    def shots(gates: List[Operation],
              n_qubits,
              shots: int,
              initial: Optional[np.ndarray] = None) -> typing.Counter[str]:
        if initial is not None:
            initial = _check_and_transform_initial(initial, n_qubits)
        ctx = _NumPyBackendContext(n_qubits)
        for _ in range(shots):
            ctx = NumPyBackend._run_inner(ctx, gates, n_qubits, initial)
            if ctx.cregs:
                ctx.store_shot()
        return ctx.shots_result

    @staticmethod
    def oneshot(
            gates: List[Operation],
            n_qubits,
            initial: Optional[np.ndarray] = None) -> Tuple[np.ndarray, str]:
        if initial is not None:
            initial = _check_and_transform_initial(initial, n_qubits)
        ctx = _NumPyBackendContext(n_qubits)
        ctx = NumPyBackend._run_inner(ctx, gates, n_qubits, initial)
        if ctx.cregs:
            ctx.store_shot()
        return ctx.qubits, ctx.shots_result.most_common()[0][0]

    @staticmethod
    def samples(gates: List[Operation],
              n_qubits,
              shots: int,
              initial: Optional[np.ndarray] = None) -> List[Dict[str, List[int]]]:
        """This feature is experimental."""
        if initial is not None:
            initial = _check_and_transform_initial(initial, n_qubits)
        ctx = _NumPyBackendContext(n_qubits)
        samples = []
        for _ in range(shots):
            ctx = NumPyBackend._run_inner(ctx, gates, n_qubits, initial)
            samples.append(ctx.sample)
        return samples

    def make_cache(self, gates: List[Operation], n_qubits: int) -> None:
        self.run(gates, n_qubits, save_cache=True)



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blueqat import Circuit
from ..gate import *
from .backendbase import Backend


class ManyQubitGateFallbacker(Backend):
    """Decomposite more than 2 qubit gate by fallback system."""
    def _run_inner(self, gates, n_qubits) -> List[Operation]:
        decomposed = []
        for gate in gates:
            if gate.n_qargs > 2:
                decomposed += self._run_inner(gate.fallback(n_qubits),
                                              n_qubits)
            else:
                decomposed.append(gate)
        return decomposed

    def run(self, gates, n_qubits, *args, **kwargs) -> Circuit:
        return Circuit(n_qubits, self._run_inner(gates, n_qubits))



import numpy as np
from .ordering import *
from scipy import special
from thewalrus import hafnian

def GaussianWigner(xi, V, mu):
    xi = xi - mu
    xi_tmp = np.ravel(xi)
    N = np.int(len(xi_tmp) / 2)
    det_V = np.linalg.det(V)
    V_inv = np.linalg.inv(V)
    W = (2 * np.pi)**(-N) / np.sqrt(det_V) * np.exp(-1/2 * np.dot(xi_tmp, np.dot(V_inv, xi_tmp.T)))
    return W

def StateAfterMeasurement(mu, V, idx, res, Pi):
    N = np.int(V.shape[0] / 2)
    subSysA = np.delete(np.delete(V, [2 * idx, 2 * idx + 1], 0), [2 * idx, 2 * idx + 1], 1)
    subSysB = V[(2 * idx):(2 * idx + 2), (2 * idx):(2 * idx + 2)]
    arrayList = []
    for j in range(N):
        if j != idx:
            arrayList.append(V[(2 * j):(2 * j + 2), (2 * idx):(2 * idx + 2)])
    C = np.concatenate(arrayList)
    post_V = subSysA - np.dot(C, np.dot(1 / np.sum(subSysB * Pi) * Pi, C.T))
    post_V = np.insert(post_V, 2 * idx, [[0], [0]], axis = 0)
    post_V = np.insert(post_V, 2 * idx, [[0], [0]], axis = 1)
    post_V[2 * idx, 2 * idx] = 1
    post_V[2 * idx + 1, 2 * idx + 1] = 1
    
    post_mu = np.delete(mu, [2 * idx, 2 * idx + 1]) + \
    np.dot(np.dot(C, 1 / np.sum(subSysB * Pi) * Pi), res * np.diag(Pi) - mu[(2 * idx):(2 * idx + 2)])
    post_mu = np.insert(post_mu, 2 * idx, [0, 0])
    
    return post_mu, post_V

def GaussianQfunc(alpha, V, mu):
    mu_Q = RtoTvec(mu)
    V_Q = RtoTmat(V)
    alpha_Q = RtoTvec(alpha)
    alpha_Q = alpha_Q - mu_Q
    V_Q = V_Q + (np.eye(V_Q.shape[0]) * 0.5)

    det_V_Q = np.linalg.det(V_Q)
    V_Qinv = np.linalg.inv(V_Q)

    Q = 1 / np.sqrt(det_V_Q * np.pi) * np.exp(-1/2 * np.dot(np.conj(alpha_Q), np.dot(V_Qinv, alpha_Q)))

def FockDensityMatrix(cov, mu, m, n, tol = 1e-10):
    if np.max(cov - cov.T) > tol:
        raise ValueError("Covariance matrix must be symmetric.")
    else:
        cov = (cov + cov.T) / 2

    N = np.int(cov.shape[0] / 2)
    cov_Q = RtoQmat(cov)
    cov_Q_inv = np.linalg.inv(cov_Q)
    cov_Q_det = np.linalg.det(cov_Q)
    mu_Q = RtoTvec(mu)

    T1 = np.exp(-0.5 * np.dot(np.dot(np.conj(mu_Q), cov_Q_inv), mu_Q))
    T2 = np.sqrt(cov_Q_det * np.prod(special.factorial(m)) * np.prod(special.factorial(n)))
    T = T1 / T2

    X = np.block([[np.zeros([N, N]), np.eye(N)], [np.eye(N), np.zeros([N, N])]])
    A = np.dot(X, (np.eye(2 * N) - cov_Q_inv))
    A = (A + A.T) / 2 # cancel the numeric error of symmetray
    A_rp = np.repeat(A, np.hstack([n, m]), axis = 0)
    A_rp = np.repeat(A_rp, np.hstack([n, m]), axis = 1)
    gamma = np.dot(np.conj(mu_Q), cov_Q_inv)
    gamma_rp = np.repeat(gamma, np.hstack([n, m]))
    np.fill_diagonal(A_rp, gamma_rp)
    prob = T * hafnian(A_rp, loop = True)
    return prob




import numpy as np
from scipy.special import factorial as fact

def coherent(initState, mode, alpha, cutoff):
    n = np.arange(cutoff + 1)
    state = np.exp(- 0.5 * np.abs(alpha) ** 2) / np.sqrt(fact([n])) * alpha ** n
    initState[mode, :] = state
    return initState

def vacuum(initState, mode):
    initState[mode, :] = 0
    initState[mode, 0] = 1
    return initState

# arXiv:quant-ph/0509137
def cat(initState, mode, alpha, parity, cutoff):
    n = np.arange(cutoff + 1)
    if parity == 'e':
        N = 1 / np.sqrt(2 * (1 + np.exp(-2 * np.abs(alpha) ** 2)))
        coeff = 2 * N * np.exp(-(np.abs(alpha) ** 2) / 2)
        initState[mode, :] = coeff * alpha ** (n) / np.sqrt(fact(n)) * np.mod(n + 1, 2)
    elif parity == 'o':
        N = 1 / np.sqrt(2 * (1 - np.exp(-2 * np.abs(alpha)**2)))
        coeff = 2 * N * np.exp(-(np.abs(alpha) ** 2) / 2)
        initState[mode, :] = coeff * alpha ** (n) / np.sqrt(fact(n)) * np.mod(n, 2)
    return initState

def photonNumState(initState, mode, n, cutoff):
    photonNumState = np.zeros(cutoff + 1)
    photonNumState[n] = 1
    initState[mode, :] = photonNumState
    return initState


# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Blueqat numba backend."""

import cmath
import math
import random
import warnings
from collections import Counter
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np
from numba import njit, prange
import numba

from ..gate import *
from .backendbase import Backend

DEFAULT_DTYPE = complex
FASTMATH = True

# Typedef
# Index of Qubit
_QBIdx = numba.uint32
_QBIdx_dtype = np.uint32
# Index of Quantum State
_QSIdx = numba.uint64
_QSIdx_dtype = np.uint64
# Mask of Quantum State
_QSMask = _QSIdx
_QSMask_dtype = _QSIdx_dtype


@njit(_QSMask(_QSMask, _QSIdx),
      locals={
          'lower': _QSMask,
          'higher': _QSMask
      },
      nogil=True,
      cache=True)
def _shifted(lower_mask: _QSMask, idx: _QSIdx) -> _QSMask:
    lower = idx & lower_mask
    higher = (idx & ~lower_mask) << 1
    return higher | lower


@njit(_QSMask[:](_QBIdx[:]), nogil=True, cache=True)
def _create_masks(indices: np.ndarray) -> np.ndarray:
    indices.sort()
    masks = np.empty(len(indices) + 1, dtype=_QSMask_dtype)
    for i, x in enumerate(indices):
        masks[i] = (1 << (x - i)) - 1
    masks[-1] = ~0
    for i in range(len(indices), 0, -1):
        masks[i] &= ~masks[i - 1]
    return masks


@njit(_QSMask(_QSMask[:], _QSIdx), nogil=True, cache=True)
def _mult_shifted(masks: np.ndarray, idx: _QSIdx) -> _QSMask:
    shifted = 0
    for i, x in enumerate(masks):
        shifted |= (idx & x) << i
    return shifted


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _zgate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx) -> None:
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        qubits[_shifted(lower_mask, i) + (1 << target)] *= -1


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _xgate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx) -> None:
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        qubits[i0] = qubits[i0 + (1 << target)]
        qubits[i0 + (1 << target)] = t


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _ygate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx) -> None:
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0] * 1j
        qubits[i0] = qubits[i0 + (1 << target)] * -1j
        qubits[i0 + (1 << target)] = t


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _hgate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx) -> None:
    sqrt2_inv = 0.7071067811865475
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = (t + u) * sqrt2_inv
        qubits[i0 + (1 << target)] = (t - u) * sqrt2_inv


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _diaggate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx,
              factor: np.float64) -> None:
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i1 = _shifted(lower_mask, i) + (1 << target)
        qubits[i1] *= factor


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _rzgate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx,
            ang: np.float64) -> None:
    ang *= 0.5
    eit = cmath.exp(1.j * ang)
    eitstar = eit.conjugate()
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0] *= eitstar
        qubits[i0 + (1 << target)] *= eit


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _rygate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx,
            ang: np.float64) -> None:
    ang *= 0.5
    cos = math.cos(ang)
    sin = math.sin(ang)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = cos * t - sin * u
        qubits[i0 + (1 << target)] = sin * t + cos * u


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _rxgate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx,
            ang: np.float64) -> None:
    ang *= 0.5
    cos = math.cos(ang)
    nisin = math.sin(ang) * -1.j
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = cos * t + nisin * u
        qubits[i0 + (1 << target)] = nisin * t + cos * u


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _ugate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx,
        theta: np.float64, phi: np.float64, lam: np.float64, gamma: np.float64) -> None:
    cos = math.cos(theta * 0.5)
    sin = math.sin(theta * 0.5)
    a = cos * cmath.exp(1j * gamma)
    b = -sin * cmath.exp(1j * (gamma + lam))
    c = sin * cmath.exp(1j * (gamma + phi))
    d = cos * cmath.exp(1j * (gamma + phi + lam))
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = a * t + b * u
        qubits[i0 + (1 << target)] = c * t + d * u


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _mat1gate(qubits: np.ndarray, n_qubits: _QSIdx, target: _QSIdx,
              mat: np.ndarray) -> None:
    lower_mask = (1 << _QSMask(target)) - 1
    a = mat[0, 0]
    b = mat[0, 1]
    c = mat[1, 0]
    d = mat[1, 1]
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = a * t + b * u
        qubits[i0 + (1 << target)] = c * t + d * u


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _czgate(qubits: np.ndarray, n_qubits: _QSIdx,
            controls_target: np.ndarray) -> None:
    #target = controls_target[-1]
    all1 = _QSMask(0)
    for b in controls_target:
        all1 |= _QSMask(1) << b
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i11 = _mult_shifted(masks, i) | all1
        qubits[i11] *= -1


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _cxgate(qubits: np.ndarray, n_qubits: _QSIdx,
            controls_target: np.ndarray) -> None:
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        qubits[i10] = qubits[i11]
        qubits[i11] = t


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _crxgate(qubits: np.ndarray, n_qubits: _QSIdx, controls_target: np.ndarray,
             ang: np.float64) -> None:
    ang *= 0.5
    cos = math.cos(ang)
    nisin = math.sin(ang) * -1.j
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        u = qubits[i11]
        qubits[i10] = cos * t + nisin * u
        qubits[i11] = nisin * t + cos * u


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _crygate(qubits: np.ndarray, n_qubits: _QSIdx, controls_target: np.ndarray,
             ang: np.float64) -> None:
    ang *= 0.5
    cos = math.cos(ang)
    sin = math.sin(ang)
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        u = qubits[i11]
        qubits[i10] = cos * t - sin * u
        qubits[i11] = sin * t + cos * u


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _crzgate(qubits: np.ndarray, n_qubits: _QSIdx, controls_target: np.ndarray,
             ang: np.float64) -> None:
    ang *= 0.5
    eit = cmath.exp(1.j * ang)
    eitstar = eit.conjugate()
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        qubits[i10] *= eitstar
        qubits[i11] *= eit


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _cphasegate(qubits: np.ndarray, n_qubits: _QSIdx,
                controls_target: np.ndarray, ang: np.float64) -> None:
    eit = cmath.exp(1.j * ang)
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i11 = _mult_shifted(masks, i) | c_mask | t_mask
        qubits[i11] *= eit


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _cugate(qubits: np.ndarray, n_qubits: _QSIdx, controls_target: np.ndarray,
        theta: np.float64, phi: np.float64, lam: np.float64, gamma: np.float64) -> None:
    cos = math.cos(theta * 0.5)
    sin = math.sin(theta * 0.5)
    m11 = cos * cmath.exp(1j * gamma)
    m12 = -sin * cmath.exp(1j * (gamma + lam))
    m21 = sin * cmath.exp(1j * (gamma + phi))
    m22 = cos * cmath.exp(1j * (gamma + phi + lam))
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        u = qubits[i11]
        qubits[i10] = m11 * t + m12 * u
        qubits[i11] = m21 * t + m22 * u


@njit(locals={'lower_mask': _QSMask},
      nogil=True,
      parallel=True,
      fastmath=FASTMATH)
def _p0calc(qubits: np.ndarray, target: _QSIdx,
            n_qubits: _QSIdx) -> np.float64:
    p0 = 0.0
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        val = qubits[_shifted(lower_mask, i)]
        p0 += val.real * val.real + val.imag * val.imag
    return p0


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _reduce0(qubits: np.ndarray, target: _QSIdx, n_qubits: _QSIdx,
             p0: np.float64) -> None:
    sqrtp_inv = 1.0 / math.sqrt(p0)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0] *= sqrtp_inv
        qubits[i0 + (1 << target)] = 0.0


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _reduce1(qubits: np.ndarray, target: _QSIdx, n_qubits: _QSIdx,
             p0: np.float64) -> None:
    sqrtp_inv = 1.0 / math.sqrt(1.0 - p0)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0 + (1 << target)] *= sqrtp_inv
        qubits[i0] = 0.0


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _reset1(qubits: np.ndarray, target: _QSIdx, n_qubits: _QSIdx,
            p0: np.float64) -> None:
    sqrtp_inv = 1.0 / math.sqrt(1.0 - p0)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0] = qubits[i0 + (1 << target)] * sqrtp_inv
        qubits[i0 + (1 << target)] = 0.0


class _NumbaBackendContext:
    """This class is internally used in NumbaBackend"""
    def __init__(self,
                 n_qubits: int,
                 save_cxt_cache: bool,
                 cache: Optional[np.ndarray],
                 cache_idx: int,
                 dtype=DEFAULT_DTYPE) -> None:
        self.n_qubits: int = n_qubits
        self.qubits: np.ndarray = np.zeros(2**n_qubits, dtype)
        self.save_cxt_cache: bool = save_cxt_cache
        self.shots_result: Counter = Counter()
        self.cregs: List[int] = [0] * self.n_qubits
        self.cache: Optional[np.ndarray] = cache
        self.cache_idx: int = cache_idx

    def prepare(self, initial: Optional[np.ndarray]) -> None:
        """Prepare to run next shot."""
        if self.cache is not None:
            np.copyto(self.qubits, self.cache)
        elif initial is not None:
            np.copyto(self.qubits, initial)
        else:
            self.qubits.fill(0.0)
            self.qubits[0] = 1.0
        self.cregs = [0] * self.n_qubits

    def store_shot(self) -> None:
        """Store current cregs to shots_result"""
        def to_str(cregs):
            return ''.join(str(b) for b in cregs)

        key = to_str(self.cregs)
        self.shots_result[key] = self.shots_result.get(key, 0) + 1


class NumbaBackend(Backend):
    """Simulator backend which uses numba."""
    __return_type: Dict[str, Callable[[_NumbaBackendContext], Any]] = {
        "statevector": lambda ctx: ctx.qubits,
        "shots": lambda ctx: ctx.shots_result,
        "statevector_and_shots": lambda ctx: (ctx.qubits, ctx.shots_result),
        "_inner_ctx": lambda ctx: ctx,
    }
    DEFAULT_SHOTS: int = 1024

    def __init__(self) -> None:
        self.cache = None
        self.cache_idx = -1

    def __clear_cache(self) -> None:
        self.cache = None
        self.cache_idx = -1

    def __clear_cache_if_invalid(self, n_qubits: int, dtype: type) -> None:
        if self.cache is None:
            self.__clear_cache()
            return
        if len(self.cache) != 2**n_qubits:
            self.__clear_cache()
            return
        if self.cache.dtype != dtype:
            self.__clear_cache()
            return

    def run(self,
            gates: List[Operation],
            n_qubits: int,
            shots: Optional[int] = None,
            returns: Optional[str] = None,
            initial: Optional[np.ndarray] = None,
            save_cache: bool = False,
            ignore_global: bool = False,
            dtype: type = DEFAULT_DTYPE,
            enable_ctx_cache: Optional[bool] = None,
            **kwargs) -> Any:
        def __parse_shots_returns(shots: Optional[int],
                                  returns: Optional[str]) -> Tuple[int, str]:
            if returns is None:
                if shots is None:
                    returns = "statevector"
                else:
                    returns = "shots"
            if returns not in self.__return_type.keys():
                raise ValueError(f"Unknown returns type '{returns}'")
            if shots is None:
                if returns in ("statevector", "_inner_ctx"):
                    shots = 1
                else:
                    shots = self.DEFAULT_SHOTS
            if returns == "statevector" and shots > 1:
                warnings.warn(
                    "When `returns` = 'statevector', `shots` = 1 is enough.")
            return shots, returns

        shots, returns = __parse_shots_returns(shots, returns)

        if enable_ctx_cache is None:
            enable_ctx_cache = shots > 1
        elif enable_ctx_cache is False:
            self.__clear_cache()

        if kwargs:
            warnings.warn(f"Unknown arguments {kwargs}")

        if initial is not None:
            if not isinstance(initial, np.ndarray):
                raise ValueError(f"`initial` must be a np.ndarray, but {type(initial)}")
            if initial.shape != (2**n_qubits,):
                raise ValueError(f"`initial.shape` is not matched. Expected: {(2**n_qubits,)}, Actual: {initial.shape}")
            if initial.dtype != DEFAULT_DTYPE:
                initial = initial.astype(DEFAULT_DTYPE)
            if save_cache:
                warnings.warn("When initial is not None, saving cache is disabled.")
                save_cache = False
            self.__clear_cache()

        self.__clear_cache_if_invalid(n_qubits, dtype)

        ctx = _NumbaBackendContext(n_qubits, enable_ctx_cache, self.cache,
                                   self.cache_idx, dtype)

        def run_single_gate(gate: Operation) -> None:
            nonlocal ctx
            action = self._get_action(gate)
            if action is not None:
                ctx = action(gate, ctx)
            elif isinstance(gate, IFallbackOperation):
                for g in gate.fallback(n_qubits):
                    run_single_gate(g)
            else:
                raise ValueError(f"Unknown operation {gate.lowername}.")

        for _ in range(shots):
            ctx.prepare(initial)
            cache_idx = ctx.cache_idx
            for gate in gates[cache_idx + 1:]:
                run_single_gate(gate)
                if ctx.save_cxt_cache:
                    cache_idx += 1
            if ctx.cache_idx != cache_idx:
                ctx.cache_idx = cache_idx
                if ctx.save_cxt_cache:
                    ctx.cache = ctx.qubits.copy()
                if save_cache:
                    self.cache_idx = ctx.cache_idx
                    self.cache = ctx.cache
            if ctx.cregs:
                ctx.store_shot()

        if ignore_global:
            _ignore_global(ctx.qubits)
        return self.__return_type[returns](ctx)

    def make_cache(self, gates: List[Operation], n_qubits: int) -> None:
        self.run(gates, n_qubits)

    @staticmethod
    def gate_x(gate: XGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of X gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _xgate(qubits, n_qubits, target)
        return ctx

    @staticmethod
    def gate_y(gate: YGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of Y gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _ygate(qubits, n_qubits, target)
        return ctx

    @staticmethod
    def gate_z(gate: ZGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of Z gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _zgate(qubits, n_qubits, target)
        return ctx

    @staticmethod
    def gate_h(gate: HGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of H gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _hgate(qubits, n_qubits, target)
        return ctx

    @staticmethod
    def gate_t(gate: TGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of T gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = cmath.exp(0.25j * math.pi)
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    @staticmethod
    def gate_tdg(gate: TDagGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of T† gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = cmath.exp(-0.25j * math.pi)
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    @staticmethod
    def gate_s(gate: SGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of S gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = 1.j
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    @staticmethod
    def gate_sdg(gate: SDagGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of S† gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = -1.j
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    @staticmethod
    def gate_cu(gate: CUGate, ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CU gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for control, target in gate.control_target_iter(n_qubits):
            _cugate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype),
                gate.theta, gate.phi, gate.lam, gate.gamma)
        return ctx

    @staticmethod
    def gate_cx(gate: CXGate,
                ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CX gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for control, target in gate.control_target_iter(n_qubits):
            _cxgate(qubits, n_qubits,
                    np.array([control, target], dtype=_QBIdx_dtype))
        return ctx

    @staticmethod
    def gate_cz(gate: CZGate,
                ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for control, target in gate.control_target_iter(n_qubits):
            _czgate(qubits, n_qubits,
                    np.ndarray([control, target], dtype=_QBIdx_dtype))
        return ctx

    @staticmethod
    def gate_rx(gate: RXGate,
                ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of RX gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for target in gate.target_iter(n_qubits):
            _rxgate(qubits, n_qubits, target, theta)
        return ctx

    @staticmethod
    def gate_ry(gate: RYGate,
                ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of RY gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for target in gate.target_iter(n_qubits):
            _rygate(qubits, n_qubits, target, theta)
        return ctx

    @staticmethod
    def gate_rz(gate: RZGate,
                ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of RZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for target in gate.target_iter(n_qubits):
            _rzgate(qubits, n_qubits, target, theta)
        return ctx

    @staticmethod
    def gate_phase(gate: PhaseGate,
                   ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of Phase gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = cmath.exp(1.j * gate.theta)
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    @staticmethod
    def gate_crx(gate: CRXGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CRX gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _crxgate(qubits, n_qubits,
                     np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    @staticmethod
    def gate_cry(gate: CRYGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CRY gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _crygate(qubits, n_qubits,
                     np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    @staticmethod
    def gate_crz(gate: CRZGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CRZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _crzgate(qubits, n_qubits,
                     np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    @staticmethod
    def gate_cphase(gate: CPhaseGate,
                    ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CPhase gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _cphasegate(qubits, n_qubits,
                        np.array([control, target], dtype=_QBIdx_dtype),
                        theta)
        return ctx

    @staticmethod
    def gate_ccz(gate: CCZGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of CCZ gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        _czgate(qubits, n_qubits, np.array(gate.targets, dtype=_QBIdx_dtype))
        return ctx

    @staticmethod
    def gate_ccx(gate: ToffoliGate,
                 ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of Toffoli gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        _cxgate(qubits, n_qubits, np.array(gate.targets, dtype=_QBIdx_dtype))
        return ctx

    @staticmethod
    def gate_u(gate: UGate,
                ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of U gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _ugate(qubits, n_qubits, target, gate.theta, gate.phi, gate.lam, gate.gamma)
        return ctx

    @staticmethod
    def gate_mat1(gate: Mat1Gate,
                  ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of Mat1 gate."""
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        mat = gate.matrix()
        for target in gate.target_iter(n_qubits):
            _mat1gate(qubits, n_qubits, target, mat)
        return ctx

    @staticmethod
    def gate_measure(gate: Measurement,
                     ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of measurement operation."""
        if ctx.save_cxt_cache:
            ctx.cache = ctx.qubits.copy()
        ctx.save_cxt_cache = False
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            rand = random.random()
            p0 = _p0calc(qubits, target, n_qubits)
            if rand < p0:
                _reduce0(qubits, target, n_qubits, p0)
                ctx.cregs[target] = 0
            else:
                _reduce1(qubits, target, n_qubits, p0)
                ctx.cregs[target] = 1
        return ctx

    @staticmethod
    def gate_reset(gate: Reset,
                   ctx: _NumbaBackendContext) -> _NumbaBackendContext:
        """Implementation of reset operation."""
        if ctx.save_cxt_cache:
            ctx.cache = ctx.qubits.copy()
        ctx.save_cxt_cache = False
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            rand = random.random()
            p0 = _p0calc(qubits, target, n_qubits)
            if rand < p0:
                _reduce0(qubits, target, n_qubits, p0)
            else:
                _reset1(qubits, target, n_qubits, p0)
        return ctx


@njit(nogil=True, cache=True)
def _ignore_global(qubits: np.ndarray) -> np.ndarray:
    for q in qubits:
        if abs(q) > 0.0000001:
            ang = abs(q) / q
            qubits *= ang
            break
    return qubits




"""
`state` module implements quantum state preparation.
This module is internally used.
"""

import numpy as np
from .stateOps import *

class STATE():
    """Quantum state preparating class."""
    def __init__(self, obj):
        self.obj = obj

class coherentState(STATE):
    """
    Coherent state.
    """
    def __init__(self, obj, mode, alpha):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.mode = mode
        self.alpha = alpha
        super().__init__(obj)

    def run(self, initState):
        return coherent(initState, self.mode, self.alpha, self.cutoff)

class vacuumState(STATE):
    """
    Vacuum state.
    """
    def __init__(self, obj, mode):
        self.obj = obj
        self.mode = mode
        super().__init__(obj)

    def run(self, initState):
        return vacuum(initState, self.mode)

class catState(STATE):
    """
    Cat state.
    """
    def __init__(self, obj, mode, alpha, parity):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.mode = mode
        self.alpha = alpha
        self.parity = parity
        if (parity != "e") and (parity != "o"):
            raise ValueError("parity must be 'e'(even) or 'o'(odd).")
        
    def run(self, initState):
        return cat(initState, self.mode, self.alpha, self.parity, self.cutoff)

class photonNumberState(STATE):
    """
    Photon number state.
    """
    def __init__(self, obj, mode, n):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.mode = mode
        self.n = n
    def run(self, initState):
        return photonNumState(initState, self.mode, self.n, self.cutoff)



from .numpy_backend import NumPyBackend
from .numba_backend import numba_backend_lazy
from .qasm_output_backend import QasmOutputBackend
from .ibmq_backend import ibmq_backend
from .sympy_backend import SympyBackend
from .onequbitgate_transpiler import OneQubitGateCompactionTranspiler
from .twoqubitgate_transpiler import TwoQubitGateDecomposingTranspiler
from .draw_backend import DrawCircuit
from .quimb import Quimb

BACKENDS = {
    "numpy": NumPyBackend,
    "numba": numba_backend_lazy,
    "qasm_output": QasmOutputBackend,
    "ibmq": ibmq_backend,
    "sympy_unitary": SympyBackend,
    "2q_decomposition": TwoQubitGateDecomposingTranspiler,
    "1q_compaction": OneQubitGateCompactionTranspiler,
    "draw": DrawCircuit,
    "quimb": Quimb,
}
DEFAULT_BACKEND_NAME = "numpy"



import numpy as np

from blueqat import Circuit

def circuit_to_unitary(circ: Circuit, *runargs, **runkwargs) -> np.ndarray:
    """Make circuit to unitary. This function is experimental feature and
    may changed or deleted in the future."""
    runkwargs.setdefault('returns', 'statevector')
    runkwargs.setdefault('ignore_global', False)
    n_qubits = circ.n_qubits
    vecs = []
    if n_qubits == 0:
        return np.array([[1]])
    for i in range(1 << n_qubits):
        bitmask = tuple(k for k in range(n_qubits) if (1 << k) & i)
        c = Circuit()
        if bitmask:
            c.x[bitmask]
        c += circ
        vecs.append(c.run(*runargs, **runkwargs))
    return np.array(vecs).T



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for convenient."""
import cmath
from collections import Counter
import math
import typing
from typing import Dict, Iterator, Tuple, Union
import warnings

import numpy as np

if typing.TYPE_CHECKING:
    from . import Circuit


def to_inttuple(
    bitstr: Union[str, Counter, Dict[str, int]]
) -> Union[Tuple, Counter, Dict[Tuple, int]]:
    """Convert from bit string likes '01011' to int tuple likes (0, 1, 0, 1, 1)

    Args:
        bitstr (str, Counter, dict): String which is written in "0" or "1".
            If all keys are bitstr, Counter or dict are also can be converted by this function.

    Returns:
        tuple of int, Counter, dict: Converted bits.
            If bitstr is Counter or dict, returns the Counter or dict
            which contains {converted key: original value}.

    Raises:
        ValueError: If bitstr type is unexpected or bitstr contains illegal character.
    """
    if isinstance(bitstr, str):
        return tuple(int(b) for b in bitstr)
    if isinstance(bitstr, Counter):
        return Counter(
            {tuple(int(b) for b in k): v
             for k, v in bitstr.items()})
    if isinstance(bitstr, dict):
        return {tuple(int(b) for b in k): v for k, v in bitstr.items()}
    raise ValueError("bitstr type shall be `str`, `Counter` or `dict`")


def ignore_global_phase(statevec: np.ndarray) -> np.ndarray:
    """Multiple e^-iθ to `statevec` where θ is a phase of first non-zero element.

    Args:
        statevec np.ndarray: Statevector.

    Returns:
        np.ndarray: `statevec` is returned.
    """
    for q in statevec:
        if abs(q) > 0.0000001:
            ang = abs(q) / q
            statevec *= ang
            break
    return statevec


def check_unitarity(mat: np.ndarray) -> bool:
    """Check whether mat is a unitary matrix."""
    shape = mat.shape
    if len(shape) != 2 or shape[0] != shape[1]:
        # Not a square matrix.
        return False
    return np.allclose(mat @ mat.T.conjugate(), np.eye(shape[0]))


def circuit_to_unitary(circ: 'Circuit', *runargs, **runkwargs) -> np.ndarray:
    """Make circuit to unitary. This function is experimental feature and
    may changed or deleted in the future."""
    warnings.warn(
        "blueqat.util.circuit_to_unitary is moved to " +
        "blueqat.circuit_funcs.circuit_to_unitary.circuit_to_unitary.",
        DeprecationWarning)
    from blueqat.circuit_funcs.circuit_to_unitary import circuit_to_unitary as f
    return f(circ, *runargs, **runkwargs)


def calc_u_params(mat: np.ndarray) -> Tuple[float, float, float, float]:
    """Calculate U-gate parameters from a 2x2 unitary matrix."""
    assert mat.shape == (2, 2)
    assert check_unitarity(mat)
    gamma = cmath.phase(mat[0, 0])
    mat = mat * cmath.exp(-1j * gamma)
    theta = math.atan2(abs(mat[1, 0]), mat[0, 0].real) * 2.0
    phi_plus_lambda = cmath.phase(mat[1, 1])
    phi = cmath.phase(mat[1, 0]) % (2.0 * math.pi)
    lam = (phi_plus_lambda - phi) % (2.0 * math.pi)
    return theta, phi, lam, gamma


def sqrt_2x2_matrix(mat: np.ndarray) -> np.ndarray:
    """Returns square root of 2x2 matrix.

    Reference: https://en.wikipedia.org/wiki/Square_root_of_a_2_by_2_matrix
    """
    assert mat.shape == (2, 2)
    s = np.sqrt(np.linalg.det(mat))
    t = np.sqrt(mat[0, 0] + mat[1, 1] + 2 * s)
    if abs(t) < 1e-8: # Avoid zero division.
        s = -s
        t = np.sqrt(mat[0, 0] + mat[1, 1] + 2 * s)
    return (mat + s * np.eye(2)) / t


def gen_graycode(n: int) -> Iterator[int]:
    """Generate an iterator which returns Gray code."""
    return (v ^ (v >> 1) for v in range(2**n))


def gen_gray_controls(n: int) -> Iterator[Tuple[int, int, int]]:
    """Generate an iterator which returns bit indices for constructing
    Gray code based controlled gate.

    ## Example

    4 controlled Z impletation:

    ```py
    n = 3
    t = 3
    c = Circuit()
    angles = [pi / 2**(n - 1), -pi / 2**(n - 1)]
    for c0, c1, parity in gen_gray_controls(n):
        if c0 >= 0:
            c.cx[c0, c1]
        c.crz(angles[parity])[c1, t]
    """
    def gen_changedbit(n):
        pow2 = [2 ** i for i in range(n)]
        gen = gen_graycode(n)
        try:
            prev = next(gen)
        except StopIteration: # Unreachable.
            raise ValueError() from None
        for g in gen:
            yield pow2.index(g ^ prev)
            prev = g

    def gen_cxtarget():
        k = 0
        while 1:
            for _ in range(2**k):
                yield k
            k += 1

    def gen_parity():
        while 1:
            yield 0
            yield 1

    for c0, c1, p in zip(gen_changedbit(n), gen_cxtarget(), gen_parity()):
        if c0 == c1:
            yield c0 - 1, c1, p
        else:
            yield c0, c1, p



import numpy as np
from .Gaussianformula.baseFunc import *
from .Gaussianformula.ordering import *
from .Gaussianformula.gates import *
import matplotlib.pyplot as plt

GATE_SET = {
    "D": Dgate,
    "BS": BSgate,
    "S": Sgate,
    "R": Rgate,
    "XS": Sgate,
    "PS": PSgate,
    "X": Xgate,
    "Z": Zgate,
    "TMS": TMSgate,
    "MeasX": MeasX,
    "MeasP": MeasP
}

class Gaussian():
    """
    Class for continuous variable quantum compting in Gaussian formula.
    This class can only deal with gaussian states and gaussian gate.
    """    
    def __init__(self, N):
        self.N = N
        self.V = (np.eye(2 * N)) * 0.5
        self.mu = np.zeros(2 * N)
        self.ops = []
        self.creg = [[None, None] for i in range(self.N)] # [x, p]
    
    def  __getattr__(self, name):
        if name in GATE_SET:
            self.ops.append(GATE_SET[name])
            return self._setGateParam
        else:
            raise AttributeError('The state method does not exist')

    def _setGateParam(self, *args, **kwargs):
        self.ops[-1] = self.ops[-1](self, *args, **kwargs)
        return self

    def Creg(self, idx, var, scale = 1):
        return CregReader(self.creg, idx, var, scale)
    
    def run(self):
        """
        Run the circuit.
        """
        for gate in self.ops:
            [self.mu, self.V] = gate.run(state = [self.mu, self.V])
        return self

    def mean(self, idx):
        res = np.copy(self.mu[2 * idx:2 * idx + 2])
        return res

    def cov(self, idx):
        res = np.copy(self.V[(2 * idx):(2 * idx + 2), (2 * idx):(2 * idx + 2)])
        return res

    def Wigner(self, idx, plot = 'y', xrange = 5.0, prange = 5.0):
        """
        Calculate the Wigner function of a selected mode.
        
        Args:
            mode (int): Selecting a optical mode
            plot: If 'y', the plot of wigner function is output using matplotlib. If 'n', only the meshed values are returned
            x(p)range: The range in phase space for calculateing Wigner function
            
        """
        idx = idx * 2
        x = np.arange(-xrange, xrange, 0.1)
        p = np.arange(-prange, prange, 0.1)
        m = len(x)
        xx, pp = np.meshgrid(x, p)
        xi_array = np.dstack((pp, xx))
        W = np.zeros((m, m))
        for i in range(m):
            for j in range(m):
                W[i][j] = GaussianWigner(xi_array[j][i], self.V[idx:idx+2, idx:idx+2], self.mu[idx:idx+2])
        if plot == 'y':
            h = plt.contourf(x, p, W)
            plt.show()
        return (x, p, W)

    def PhotonDetectionProb(self, m, n):
        """
        Calculate Fock density matrix element rho_{mn}.
        m and n should be numpy array which length is same as mode number.
        When m = n, the returned value is probability for photon number m is measured.
        For example, if m = n = np.array([0, 0]), returned value is probability \
            for detecting 0 photon for both two modes.
        """
        if len(m) != self.N or len(n) != self.N:
            raise ValueError("Input array dimension must be same as mode Number.")
        return np.real(FockDensityMatrix(self.V, self.mu, m, n))

    def GaussianToFock(self, cutoffDim = 10):
        photonNumList = []
        cutoffDim += 1
        rho = np.empty([cutoffDim ** self.N, cutoffDim ** self.N])
        for i in range(cutoffDim ** self.N):
            photonNum = []
            for j in range(self.N):
                photonNum.insert(0, np.int(np.floor(i / (cutoffDim ** j))) % cutoffDim)
            photonNumList.append(photonNum)

        for i in range(cutoffDim ** self.N):
            for j in range(cutoffDim ** self.N):
                m = np.array(photonNumList[i])
                n = np.array(photonNumList[j])
                row = [m[i] ** (self.N - i - 1) for i in range(self.N)]
                col = [n[i] ** (self.N - i - 1) for i in range(self.N)]
                rho[row, col] = FockDensityMatrix(self.V, self.mu, m, n)

        return rho

    def Interferometer(self, U):
        num = self.N
        n = U.shape[0]
        BSang = [0]*(int(n*(n-1)/2))
        rot = [0]*(int(n*(n-1)/2))
        for i in range (1, n):
            if ((i+2)%2==1):
                for j in range (i): 
                    T = np.identity(n, dtype = complex)
                    if U[n-j-1][i-j-1]==0:
                        theta = 0
                        alpha = 0
                    elif U[n-j-1][i-j]==0:
                        theta = 0
                        alpha = np.pi/2
                    else:
                        theta = np.angle(U[n-j-1][i-1-j])-np.angle(U[n-j-1][i-j])
                        alpha = np.arctan(np.absolute(U[n-j-1][i-1-j])/np.absolute(U[n-j-1][i-j]))
                    BSang[int(i/2)+j] = alpha
                    rot[int(i/2)+j] = theta
                    e = np.cos(-theta) + np.sin(-theta)*1j
                    T[i-1-j][i-1-j] = e*(np.cos(alpha)+0*1j)
                    T[i-1-j][i-j] = -np.sin(alpha)+0*1j
                    T[i-j][i-1-j] = e*(np.sin(alpha)+0*1j)
                    T[i-j][i-j] = np.cos(alpha)+0*1j
                    U = U @ np.transpose(T)
                else:
                    for j in range (i):
                        T = np.identity(n, dtype = complex)
                        if U[n-i+j][j]==0:
                            theta = 0
                            alpha = 0
                        elif U[n-i+j-1][j]==0:
                            theta = 0
                            alpha = np.pi/2
                        else:
                            theta = -np.angle(U[n-i+j-1][j])+np.angle(U[n-i+j][j])
                            alpha = np.arctan(-(np.absolute(U[n-i+j][j])/np.absolute(U[n-i+j-1][j])))
                        BSang[-(i-1+j)] = alpha
                        rot[-(i-1+j)] = theta
                        e = np.cos(theta) + np.sin(theta)*1j
                        T[n-i+j-1][n-i+j-1] = e*(np.cos(alpha)+0*1j)
                        T[n-i+j-1][n-i+j] = -np.sin(alpha)+0*1j
                        T[n-i+j][n-i+j-1] = e*(np.sin(alpha)+0*1j)
                        T[n-i+j][n-i+j] = np.cos(alpha)+0*1j
                        U = T @ U

        counter = 0
        for i in range(1, n, 2):
            for k in range(i):
                self.ops.append(GATE_SET['R'])
                self.ops[-1] = self.ops[-1](self, i-1-k, rot[counter])
                self.ops.append(GATE_SET['BS'])
                self.ops[-1] = self.ops[-1](self, i-1-k, i-1-k+1, BSang[counter])
                counter += 1
        for i in reversed(range(2, n-2 if (n-1)%2==0 else n-1, 2)):
            for k in range(i):
                self.ops.append(GATE_SET['R'])
                self.ops[-1] = self.ops[-1](self, n-2-k, rot[counter])
                self.ops.append(GATE_SET['BS'])
                self.ops[-1] = self.ops[-1](self, n-2-k, n-1-k, BSang[counter])
                counter += 1




# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Decorators."""
from keyword import iskeyword
from typing import Callable, Union

from .circuit import BlueqatGlobalSetting


def circuitmacro(func: Union[Callable, str, None] = None,
                 *,
                 allow_overwrite: bool = True) -> Callable:
    """@circuitmacro decorator.

    Typical usage:
    Case 1: no arguments

        @def_macro
        def egg(c):
            ...

    equivalent to this:

        def egg(c):
            ...
        BlueqatGlobalSetting.register_macro('egg', egg, allow_overwrite=True)


    Case 2: with name:

        @def_macro('bacon')
        def egg(c):
            ...

    is equivalent with

        def egg(c):
            ...
        BlueqatGlobalSetting.register_macro('bacon', egg, allow_overwrite=True)

    Case 3: with allow_overwrite keyword argument

        @def_macro(allow_overwrite=False)
        def egg(c):
            ...

    or

        @def_macro('bacon', allow_overwrite=False)
        def bacon(c):
            ...

    call BlueqatGlobalSetting.register_macro with allow_overwrite=False.

    Please note that `allow_overwrite=True` is default behavior.
    It is convenient for interactive environment likes Jupyter Notebook.
    However, if you're library developer, using `allow_overwrite=False` is hardly recommended.
    """
    if callable(func):
        # @def_macro pattern.
        name = func.__name__
        if not name.isidentifier() or iskeyword(name):
            raise ValueError(
                f'Function name {name} is not a valid macro name. ')
        BlueqatGlobalSetting.register_macro(name, func, allow_overwrite)
        return func
    if isinstance(func, str):
        # @def_macro(name) or @def_macro(name, allow_overwrite) pattern.
        name = func

        def _wrapper1(func):
            BlueqatGlobalSetting.register_macro(name, func, allow_overwrite)
            return func

        return _wrapper1
    if func is None:
        # @def_macro(allow_overwrite) pattern.
        def _wrapper2(func):
            name = func.__name__
            if not name.isidentifier() or iskeyword(name):
                raise ValueError(
                    f'Function name {name} is not a valid macro name. ')
            BlueqatGlobalSetting.register_macro(name, func, allow_overwrite)
            return func

        return _wrapper2
    raise TypeError('Invalid type for first argument.')



"""This module provides for type hinting."""

import typing
from typing import Generic, TypeVar, Union

if typing.TYPE_CHECKING:
    from blueqat import Circuit

T = TypeVar('T')
Targets = Union[int, slice, tuple]

class CircuitOperation(Generic[T]):
    """Type definition of dynamic method."""
    def __call__(self, *args) -> T:
        ...

    def __getitem__(self, targets: Targets) -> 'Circuit':
        ...




"""
`gateOps` module implements calculation for quantum gate operations.
This module may be redefined as a backend option in later versions.
"""


import numpy as np
from .bosonicLadder import *
from .WignerFunc import reduceState, partialTrace
from scipy.special import eval_hermite as hermite
from scipy.linalg import expm

def Displacement(state, mode, alpha, modeNum, cutoff):
    state = _singleGate_preProcess(state, mode)
    state = exp_annihiration(state, -np.conj(alpha), cutoff = cutoff)
    state = exp_creation(state, alpha, cutoff = cutoff)
    state = _singleGate_postProcess(state, mode, modeNum)
    state = state * np.exp(-np.abs(alpha)**2 / 2)
    return state

def Beamsplitter(state, mode1, mode2, theta, modeNum, cutoff):
    if modeNum < 2:
        raise ValueError("The gate requires more than one mode.")
    state = _twoModeGate_preProcess(state, mode1, mode2)
    state = exp_BS(state, -theta, cutoff = cutoff)
    state = _twoModeGate_postProcess(state, mode1, mode2, modeNum)
    return state

def Squeeze(state, mode, r, phi, modeNum, cutoff):
    G = np.exp(2 * 1j * phi) * np.tanh(r)
    state = _singleGate_preProcess(state, mode)
    state = exp_annihiration(state, np.conj(G) / 2, order = 2, cutoff = cutoff)
    state = exp_photonNum(state, -np.log(np.cosh(r)), cutoff = cutoff)
    state = exp_creation(state, -G / 2, order = 2, cutoff = cutoff)
    state = _singleGate_postProcess(state, mode, modeNum)
    state = state / np.sqrt(np.cosh(r))
    return state

# def Squeeze(state, mode, r, phi, modeNum, cutoff):
#     state = _singleGate_preProcess(state, mode)
#     state = exp_aa_minus_AA(state, r * np.exp(2j * phi) / 2, cutoff = cutoff)
#     state = _singleGate_postProcess(state, mode, modeNum)
#     return state

def KerrEffect(state, mode, chi, modeNum, cutoff):
    state = _singleGate_preProcess(state, mode)
    state = exp_AAaa(state, 1j * chi / 2, cutoff = cutoff)
    state = _singleGate_postProcess(state, mode, modeNum)
    return state

def HamiltonianEvo(state, mode, expr, gamma, modeNum, cutoff):
    dim = state.shape[0]
    str_aA = expand_xp_to_aA(expr, dim, evaluate = False, hbar = 1)
    str_exp_aA = exp_str_aA(str_aA, -1j * gamma, N = 5)
    mat = str_to_aA_mat(str_exp_aA, dim, hbar = 1)
    state = _singleGate_preProcess(state, mode)
    mat_ = expm(-1j * gamma * mat)
    state = np.dot(mat_, state.T)
    state = _singleGate_postProcess(state.T, mode, modeNum)
    return state

def photonMeasurement(state, mode, post_select):
    reducedDensity = reduceState(state, mode)
    probs = np.real(np.diag(reducedDensity))
    probs = probs / np.sum(probs)
    if post_select is None:
        res = np.random.choice(probs.shape[0], 1, p = probs)
    else:
        res = post_select
    prob = probs[res]
    
    state_ = np.swapaxes(state, mode, 0)
    ind = np.ones((state.shape[-1]), bool)
    ind[res] = False
    state_[ind] = 0
    state_ = np.swapaxes(state_, mode, 0)
    state_ = state_ / np.sqrt(prob)
    return res, state_

def homodyneMeasurement(state, mode, theta, post_select):
    if post_select is None:
        res, psi = homodyneFock(state, mode, theta, ite = 1, hbar = 1)
    else:
        res = post_select   
    state_ = _afterHomodyne(state, mode, res, psi, theta)
    return res, state_

def homodyneFock(state, mode, theta, ite = 1, hbar = 1):
    dim = state.shape[0]
    rho = reduceState(state, mode)
    lim = np.sqrt(dim * 2) + 3 # heuristic
    q = np.arange(-lim, lim, lim / 500)
    q_ = q / np.sqrt(hbar)
    psi = _positionWaveFunc(dim, q_, theta) # psi_{n}(q, theta)
    probs = np.zeros(q.shape) + 0j
    for i in range(dim):
        for j in range(dim):
            tmp = rho[i, j] * psi[i, :] * np.conj(psi[j, :])# * np.exp(1j * theta * (j - i))
            probs += tmp
    probs = np.abs(np.real(probs))
    probs_sum = np.abs(np.sum(probs))
    if 1 - probs_sum < 0.01:
        probs = probs / probs_sum
    else:
        print("sum of probability", probs_sum)
        raise ValueError('probabilities do not sum to 1')
    res = np.random.choice(probs.shape[0], ite, p = probs)
    return q[res], psi[:, res]

def _afterHomodyne(state, mode, q, psi, theta):
    dim = state.shape[-1]
    modeNum = state.ndim
    #psi = _positionWaveFunc(dim, q, theta)
    state = _singleGate_preProcess(state, mode)
    state = state * psi.T
    state[:, 0] = np.sum(state, axis = -1)
    state[:, 1:] = 0
    state = _singleGate_postProcess(state, mode, modeNum)
    state = state / np.sqrt(np.sum(np.abs(state) ** 2))
    return state

def _positionWaveFunc(n, q, theta, hbar = 1):
    #lim = np.sqrt(n * 2) + 3 # heuristic
    n_ = np.arange(n)[:, np.newaxis]
    H = hermite(n_, q)
    A = (np.pi * hbar) ** 0.25 / np.sqrt(2 ** n_ * fact(n))
    psi = A * H * np.exp(-q ** 2 / 2 / hbar) * np.exp(-1j * n_ * theta)
    C = np.sum(np.abs(psi) ** 2, axis=1) + 0j # sum of squeres
    psi = psi / np.sqrt(C)[:, np.newaxis] # normalized wave functions
    return psi

def _singleGate_preProcess(fockState, mode):
    cutoff = fockState.shape[-1] - 1
    fockState = np.swapaxes(fockState, mode, fockState.ndim - 1)
    return fockState.reshape(-1, cutoff + 1)

def _singleGate_postProcess(fockState, mode, modeNum):
    cutoff = fockState.shape[-1] - 1
    fockState = fockState.reshape([cutoff + 1] * modeNum)
    return np.swapaxes(fockState, mode, modeNum - 1)

def _twoModeGate_preProcess(fockState, mode1, mode2):
    cutoff = fockState.shape[-1] - 1
    modeNum = fockState.ndim
    fockState = np.swapaxes(fockState, mode2, modeNum - 1)
    fockState = np.swapaxes(fockState, mode1, modeNum - 2)
    return fockState.reshape(-1, (cutoff + 1) ** 2)

def _twoModeGate_postProcess(fockState, mode1, mode2, modeNum):
    dim = np.int(np.sqrt(fockState.shape[-1]))
    fockState = fockState.reshape([dim] * modeNum)
    fockState = np.swapaxes(fockState, mode1, modeNum - 2)
    fockState = np.swapaxes(fockState, mode2, modeNum - 1)
    return fockState




import numpy as np
from .baseFunc import StateAfterMeasurement

def Xsqueeze(state, N, idx, r):
    mu = state[0]
    V = state[1]
    idx = 2 * idx
    S = np.eye(2 * N)
    S[idx:idx+2, idx:idx+2] = np.array([[np.exp(-r), 0], [0, np.exp(r)]])
    V_ = np.dot(S, np.dot(V, S.T))
    mu_ = np.dot(S, mu)
    return [mu_, V_]

def Rotation(state, N, idx, theta):
    mu = state[0]
    V = state[1]
    idx = 2 * idx
    S = np.eye(2 * N)
    S[idx:idx+2, idx:idx+2] = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    V_ = np.dot(S, np.dot(V, S.T))
    mu_ = np.dot(S, mu)
    return [mu_, V_]

# 10.1103/RevModPhys.77.513
def Beamsplitter(state, N, idx1, idx2, theta):
    mu = state[0]
    V = state[1]
    idx1 = 2 * idx1
    idx2 = 2 * idx2
    S = np.eye(2 * N)
    S[idx1:idx1+2, idx1:idx1+2] = np.array([[np.cos(theta), 0], [0, np.cos(theta)]])
    S[idx1:idx1+2, idx2:idx2+2] = np.array([[np.sin(theta), 0], [0, np.sin(theta)]])
    S[idx2:idx2+2, idx1:idx1+2] = np.array([[-np.sin(theta), 0], [0, -np.sin(theta)]])
    S[idx2:idx2+2, idx2:idx2+2] = np.array([[np.cos(theta), 0], [0, np.cos(theta)]])
    V_ = np.dot(S, np.dot(V, S.T))
    mu_ = np.dot(S, mu)
    return [mu_, V_]

def Displace(state, idx, alpha):
    mu = state[0]
    V = state[1]
    dx = np.real(alpha) * np.sqrt(2) # np.sqrt(2 * hbar)
    dp = np.imag(alpha) * np.sqrt(2) # np.sqrt(2 * hbar)
    mu[2 * idx:2 * idx + 2] = mu[2 * idx:2 * idx + 2] + np.array([dx, dp])
    mu_ = mu
    return [mu_, V]

def Xtrans(state, idx, dx):
    mu = state[0]
    V = state[1]
    mu[2 * idx] += dx
    mu_ = mu
    return [mu_, V]

def Ztrans(state, idx, dp):
    mu = state[0]
    V = state[1]
    mu[2 * idx + 1] += dp
    mu_ = mu
    return [mu_, V]

def twoModeSqueezing(state, N, idx1, idx2, r):
    mu = state[0]
    V = state[1]
    idx1 = 2 * idx1
    idx2 = 2 * idx2
    S = np.eye(2 * N)
    S[idx1:idx1+2, idx1:idx1+2] = np.array([[np.cosh(r), 0], [0, np.cosh(r)]])
    S[idx1:idx1+2, idx2:idx2+2] = np.array([[np.sinh(r), 0], [0, -np.sinh(r)]])
    S[idx2:idx2+2, idx1:idx1+2] = np.array([[np.sinh(r), 0], [0, -np.sinh(r)]])
    S[idx2:idx2+2, idx2:idx2+2] = np.array([[np.cosh(r), 0], [0, np.cosh(r)]])
    V_ = np.dot(S, np.dot(V, S.T))
    mu_ = np.dot(S, mu)
    return [mu_, V_]

def HomodyneX(state, idx):
    mu = state[0]
    V = state[1]
    res = np.random.normal(mu[2 * idx], np.sqrt(V[2 * idx, 2 * idx]))
    mu_, V_ = StateAfterMeasurement(mu, V, idx, res, np.diag([1, 0]))
    return res, [mu_, V_]

def HomodyneP(state, idx):
    mu = state[0]
    V = state[1]
    res = np.random.normal(mu[2 * idx + 1], np.sqrt(V[2 * idx + 1, 2 * idx + 1]))
    mu_, V_ = StateAfterMeasurement(mu, V, idx, res, np.diag([0, -1]))
    return res, [mu_, V_]





from .backendbase import Backend
from ..circuit import Circuit
from ..gate import *

# To avoid ImportError, don't import this here.
# import networkx as nx
# import matplotlib.pyplot as plt
import numpy as np
import math

class DrawCircuit(Backend):
    """Backend for draw output."""
    
    def _preprocess_run(self, gates, n_qubits, args, kwargs):
        # Lazy import to avoid unneeded ImportError.

        qlist = {}
        flg = 0
        time = 0
        add_edge = []
        remove_edge = []

        for i in range(n_qubits):
            qlist[i] = [{'num': flg, 'gate': 'q'+str(i), 'angle': '', 'xpos': 0, 'ypos': i, 'type': 'qubit'}]
            flg += 1
        
        time += 1
        return gates, (qlist, n_qubits, [flg], [time], add_edge, remove_edge)

    def _postprocess_run(self, ctx):
        import networkx as nx
        import matplotlib.pyplot as plt
        
        #color_code
        color_gate = {}
        color_gate['X'] = color_gate['Y'] = color_gate['Z'] = '#0BB0E2'
        color_gate['RX'] = color_gate['RY'] = color_gate['RZ'] = '#FCD000'
        color_gate['H'] = color_gate['T'] = color_gate['S'] = '#E6000A'
        color_gate['M'] = 'white'
        
        qlist = ctx[0]
        n_qubits = ctx[1]
        flg = ctx[2][-1]
        time = ctx[3][-1]
        
        #measurement
        for i in range(n_qubits):
            qlist[i].append({'num': flg, 'gate': 'M', 'angle': '', 'xpos': 30, 'ypos': i + math.floor((time-1)/30)*(n_qubits+1), 'type': 'measurement'})
            flg += 1
        
        G = nx.Graph()

        for i in range(n_qubits):
            for j in range(len(qlist[i])-1):
                G.add_edge(qlist[i][j]['num'], qlist[i][j+1]['num'])
        
        #twoqubit connections
        for item in ctx[4]:
            G.add_edge(item[0], item[1])

        for item in ctx[5]:
            G.remove_edge(item[0], item[1])

        #image size
        plt.figure(1, figsize=(30, (n_qubits+1)*(math.floor(time/30)+1)), dpi=60)

        labels = {}
        colors = {}
        angles = {}
        sizes = {}

        for i in range(n_qubits):
            for j in range(len(qlist[i])):
                angles[qlist[i][j]['num']] = qlist[i][j]['angle']
                labels[qlist[i][j]['num']] = qlist[i][j]['gate']
                sizes[qlist[i][j]['num']] = 1200
                if qlist[i][j]['type'] == 'dummy':
                    colors[qlist[i][j]['num']] = 'white'
                    sizes[qlist[i][j]['num']] = 0
                elif qlist[i][j]['gate'] == '' or qlist[i][j]['gate'] == 'CZ':
                    colors[qlist[i][j]['num']] = 'black'
                    sizes[qlist[i][j]['num']] = 100
                elif qlist[i][j]['type'] == 'qubit':
                    colors[qlist[i][j]['num']] = 'white'
                else:
                    colors[qlist[i][j]['num']] = color_gate[qlist[i][j]['gate']]

        
        #position
        pos = {}
        for i in range(n_qubits):
            for j in range(len(qlist[i])):
                pos[qlist[i][j]['num']] = (qlist[i][j]['xpos'], (n_qubits+1)*(math.floor(time/30)+1) - qlist[i][j]['ypos'])

        #dummy qubit just for top and bottom margin
        labels[flg]= ''
        colors[flg] = 'black'
        sizes[flg] = 0
        pos[flg] = (0, (n_qubits+1)*(math.floor(time/30)+1)+1)
        G.add_node(flg)
        labels[flg+1]= ''
        colors[flg+1] = 'black'
        sizes[flg+1] = 0
        pos[flg+1] = (0, 1)
        G.add_node(flg+1)
       
        nx.set_node_attributes(G, labels, 'label')
        nx.set_node_attributes(G, colors, 'color')
        nx.set_node_attributes(G, angles, 'angle')
        nx.set_node_attributes(G, sizes, 'size')

        options = {
            "font_size": 12,
            "edgecolors": "black",
            "linewidths": 2,
            "width": 2,
        }

        node_labels = nx.get_node_attributes(G, 'label')
        node_colors = [colors[i] for i in nx.get_node_attributes(G, 'color')]
        node_sizes = [sizes[i] for i in nx.get_node_attributes(G, 'size')]
        nx.draw_networkx(G, pos, labels = node_labels, node_color = node_colors, node_size = node_sizes, **options)

        #label positions for angles
        pos_attrs = pos.copy()
        for i in pos_attrs:
            pos_attrs[i] = (pos_attrs[i][0]+0.4, pos_attrs[i][1]-0.4)
    
        node_attrs = nx.get_node_attributes(G, 'angle')
        custom_node_attrs = {}

        for node, attr in node_attrs.items():
            custom_node_attrs[node] = attr

        nx.draw_networkx_labels(G, pos_attrs, labels = custom_node_attrs, font_size=9)
        #plt.axis('off')
        plt.show()
        
        return 

    def _one_qubit_gate_noargs(self, gate, ctx):
        flg = ctx[2][-1]
        time = ctx[3][-1]
        qlist = ctx[0]
        
        time_adjust = time%30
        if time_adjust == 0:
            for i in range(ctx[1]):
                ypos_adjust = i + (math.floor(time/30)-1)*(ctx[1]+1)
                qlist[i].append({'num': flg, 'gate': '', 'angle': '', 'xpos': 30, 'ypos': ypos_adjust, 'type': 'dummy'})
                flg += 1
            time += 1
            
            for i in range(ctx[1]):
                ypos_adjust = i + math.floor(time/30)*(ctx[1]+1)
                qlist[i].append({'num': flg, 'gate': '', 'angle': '', 'xpos': 0, 'ypos': ypos_adjust, 'type': 'dummy'})
                flg += 1
                ctx[5].append((flg-1, flg-1-ctx[1]))
        
        time_adjust = time%30
        for idx in gate.target_iter(ctx[1]):
            ypos_adjust = idx + math.floor(time/30)*(ctx[1]+1)
            qlist[idx].append({'num': flg, 'gate': gate.lowername.upper(), 'angle': '', 'xpos': time_adjust, 'ypos': ypos_adjust, 'type': 'gate'})
            flg += 1
        ctx[2].append(flg)
        ctx[3].append(time+1)
        return ctx

    gate_x = _one_qubit_gate_noargs
    gate_y = _one_qubit_gate_noargs
    gate_z = _one_qubit_gate_noargs
    gate_h = _one_qubit_gate_noargs
    gate_t = _one_qubit_gate_noargs
    gate_s = _one_qubit_gate_noargs
    
    def _one_qubit_gate_args_theta(self, gate, ctx):
        flg = ctx[2][-1]
        time = ctx[3][-1]
        qlist = ctx[0]
        
        time_adjust = time%30
        if time_adjust == 0:
            for i in range(ctx[1]):
                ypos_adjust = i + (math.floor(time/30)-1)*(ctx[1]+1)
                qlist[i].append({'num': flg, 'gate': '', 'angle': '', 'xpos': 30, 'ypos': ypos_adjust, 'type': 'dummy'})
                flg += 1
            time += 1
            
            for i in range(ctx[1]):
                ypos_adjust = i + math.floor(time/30)*(ctx[1]+1)
                qlist[i].append({'num': flg, 'gate': '', 'angle': '', 'xpos': 0, 'ypos': ypos_adjust, 'type': 'dummy'})
                flg += 1
                ctx[5].append((flg-1, flg-1-ctx[1]))
        
        time_adjust = time%30
        for idx in gate.target_iter(ctx[1]):
            ypos_adjust = idx + math.floor(time/30)*(ctx[1]+1)
            qlist[idx].append({'num': flg, 'gate': gate.lowername.upper(), 'angle': round(gate.theta, 2), 'xpos': time_adjust, 'ypos': ypos_adjust, 'type': 'gate'})
            flg += 1
        ctx[2].append(flg)
        ctx[3].append(time+1)
        return ctx

    gate_rx = _one_qubit_gate_args_theta
    gate_ry = _one_qubit_gate_args_theta
    gate_rz = _one_qubit_gate_args_theta
    gate_phase = _one_qubit_gate_args_theta

    def gate_i(self, gate, ctx):
        time = ctx[3][-1]
        ctx[3].append(time+1)
        return ctx
    
    def _two_qubit_gate_noargs(self, gate, ctx):
        flg = ctx[2][-1]
        time = ctx[3][-1]
        qlist = ctx[0]
        
        tg = ''
        if gate.lowername == 'cx':
            tg = 'x'
        elif gate.lowername == 'cy':
            tg = 'y'
        elif gate.lowername == 'cz':
            tg = 'z'

        time_adjust = time%30
        if time_adjust == 0:
            for i in range(ctx[1]):
                ypos_adjust = i + (math.floor(time/30)-1)*(ctx[1]+1)
                qlist[i].append({'num': flg, 'gate': '', 'angle': '', 'xpos': 30, 'ypos': ypos_adjust, 'type': 'dummy'})
                flg += 1
            time += 1
            
            for i in range(ctx[1]):
                ypos_adjust = i + math.floor(time/30)*(ctx[1]+1)
                qlist[i].append({'num': flg, 'gate': '', 'angle': '', 'xpos': 0, 'ypos': ypos_adjust, 'type': 'dummy'})
                flg += 1
                ctx[5].append((flg-1, flg-1-ctx[1]))

        time_adjust = time%30        
        for control, target in gate.control_target_iter(ctx[1]):
            qlist[target].append({'num': flg, 'gate': tg.upper(), 'angle': '', 'xpos': time_adjust, 'ypos': target + math.floor(time/30)*(ctx[1]+1), 'type': 'gate'})
            flg += 1
            qlist[control].append({'num': flg, 'gate': '', 'angle': '', 'xpos': time_adjust, 'ypos': control + math.floor(time/30)*(ctx[1]+1), 'type': 'gate'})
            flg += 1
            ctx[4].append((flg-2, flg-1))
        ctx[2].append(flg)
        ctx[3].append(time+1)
        return ctx
    
    gate_cx = gate_cy = gate_cz = _two_qubit_gate_noargs

    def _three_qubit_gate_noargs(self, gate, ctx):
        return ctx

    gate_ccx = _three_qubit_gate_noargs
    gate_cswap = _three_qubit_gate_noargs

    def gate_measure(self, gate, ctx):
        return ctx

    gate_reset = _one_qubit_gate_noargs



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..gate import *
from .backendbase import Backend


class QasmOutputBackend(Backend):
    """Backend for OpenQASM output."""
    def _preprocess_run(self, gates, n_qubits, args, kwargs):
        def _parse_run_args(output_prologue=True, **_kwargs):
            return {'output_prologue': output_prologue}

        args = _parse_run_args(*args, **kwargs)
        if args['output_prologue']:
            qasmlist = [
                "OPENQASM 2.0;",
                'include "qelib1.inc";',
                f"qreg q[{n_qubits}];",
                f"creg c[{n_qubits}];",
            ]
        else:
            qasmlist = []
        return gates, (qasmlist, n_qubits)

    def _postprocess_run(self, ctx):
        return "\n".join(ctx[0])

    def _one_qubit_gate_noargs(self, gate, ctx):
        for idx in gate.target_iter(ctx[1]):
            ctx[0].append(f"{gate.lowername} q[{idx}];")
        return ctx

    gate_x = _one_qubit_gate_noargs
    gate_y = _one_qubit_gate_noargs
    gate_z = _one_qubit_gate_noargs
    gate_h = _one_qubit_gate_noargs
    gate_t = _one_qubit_gate_noargs
    gate_s = _one_qubit_gate_noargs

    def _two_qubit_gate_noargs(self, gate, ctx):
        for control, target in gate.control_target_iter(ctx[1]):
            ctx[0].append(f"{gate.lowername} q[{control}],q[{target}];")
        return ctx

    gate_cz = _two_qubit_gate_noargs
    gate_cx = _two_qubit_gate_noargs
    gate_cy = _two_qubit_gate_noargs
    gate_ch = _two_qubit_gate_noargs
    gate_swap = _two_qubit_gate_noargs

    def _one_qubit_gate_args_theta(self, gate, ctx):
        for idx in gate.target_iter(ctx[1]):
            ctx[0].append(f"{gate.lowername}({gate.theta}) q[{idx}];")
        return ctx

    gate_rx = _one_qubit_gate_args_theta
    gate_ry = _one_qubit_gate_args_theta
    gate_rz = _one_qubit_gate_args_theta

    def gate_i(self, gate, ctx):
        for idx in gate.target_iter(ctx[1]):
            ctx[0].append(f"id q[{idx}];")
        return ctx

    def gate_u(self, gate, ctx):
        for idx in gate.target_iter(ctx[1]):
            if abs(gate.gamma) > 1e-7:
                ctx[0].append(
                    f"{gate.lowername}({gate.theta},{gate.phi},{gate.lam}) q[{idx}]; // global phase e^i{gate.gamma} is ignored."
                )
            else:
                ctx[0].append(
                    f"{gate.lowername}({gate.theta},{gate.phi},{gate.lam}) q[{idx}];"
                )
        return ctx

    def gate_phase(self, gate, ctx):
        for idx in gate.target_iter(ctx[1]):
            ctx[0].append(
                f"p({gate.theta}) q[{idx}];")
        return ctx

    def gate_cu(self, gate, ctx):
        for c, t in gate.control_target_iter(ctx[1]):
            ctx[0].append(
                f"{gate.lowername}({gate.theta},{gate.phi},{gate.lam},{gate.gamma}) q[{c}],q[{t}];"
            )
        return ctx

    def gate_cphase(self, gate, ctx):
        for c, t in gate.control_target_iter(ctx[1]):
            ctx[0].append(f"cp({gate.theta}) q[{c}],q[{t}];")
        return ctx

    def _three_qubit_gate_noargs(self, gate, ctx):
        c0, c1, t = gate.targets
        ctx[0].append(f"{gate.lowername} q[{c0}],q[{c1}],q[{t}];")
        return ctx

    gate_ccx = _three_qubit_gate_noargs
    gate_cswap = _three_qubit_gate_noargs

    def gate_measure(self, gate, ctx):
        for idx in gate.target_iter(ctx[1]):
            ctx[0].append(f"measure q[{idx}] -> c[{idx}];")
        return ctx

    gate_reset = _one_qubit_gate_noargs
